from abc import ABCMeta, abstractmethod


class Module(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def before_setup(self, properties, arguments, beans):
        raise NotImplementedError()

    @abstractmethod
    def after_setup(self, properties, arguments, context, beans_initializer):
        raise NotImplementedError()



import requests
import json
import framepy
import threading
import time
from framepy import core
from framepy import modules
from framepy import _thread_level_cache
from framepy import _utils

SESSION_FIELD = 'session'


class Module(modules.Module):
    def before_setup(self, properties, arguments, beans):
        app_name = _utils.resolve_property_or_report_error(
            properties=properties,
            key='app_name',
            log_message='[Eureka] Missing app_name! Skipping registration in eureka cluster'
        )
        remote_config_url = _utils.resolve_property_or_report_error(
            properties=properties,
            key='remote_config_url',
            log_message='[Eureka] Missing remote_config_url! Skipping registration in eureka cluster'
        )
        public_hostname = _utils.resolve_property_or_report_error(
            properties=properties,
            key='public_hostname',
            log_message='[Eureka] Missing public_hostname! Skipping registration in eureka cluster'
        )

        remote_config_url = self._build_eureka_url(remote_config_url)

        _register_instance(remote_config_url, app_name, public_hostname, properties)
        _register_heartbeat_service(remote_config_url, app_name, public_hostname)

        beans['_eureka_url'] = remote_config_url

    def after_setup(self, properties, arguments, context, beans_initializer):
        pass

    @staticmethod
    def _build_eureka_url(remote_config_url):
        return _utils.normalize_url(remote_config_url) + 'eureka'


def list_instances(context, service_name):
    response = _get_session_from_cache().get(context._eureka_url + '/apps/' + service_name, headers={'accept': 'application/json'})
    if response.status_code < 200 or response.status_code >= 300:
        raise Exception('Cannot retrieve instances of service ' + service_name)

    instances_list = response.json()['application']['instance']
    return [instance['hostName'] + ':' + str(instance['port']['$']) for instance in instances_list]


def _register_instance(eureka_url, app_name, hostname, properties):
    instance_data = {
        'instance': {
            'hostName': hostname,
            'ipAddr': hostname,
            'app': app_name,
            'status': 'UP',
            'port': {
                "$": properties.get('server_port', core.DEFAULT_PORT),
                "@enabled": 'true'
            },
            'dataCenterInfo': {
                "@class": "com.netflix.appinfo.InstanceInfo$DefaultDataCenterInfo",
                "name": "MyOwn"
            }
        }
    }

    try:
        response = _get_session_from_cache().post(eureka_url + '/apps/' + app_name, json.dumps(instance_data),
                                 headers={'Content-Type': 'application/json'})
        if _response_status_not_ok(response):
            framepy.log.error('[Eureka] Cannot register instance on server! Status code {0}'.format(response.status_code))
    except requests.exceptions.ConnectionError:
        framepy.log.error('[Eureka] Cannot connect to server!')


def _send_heartbeat(eureka_url, app_name, hostname):
    response = _get_session_from_cache().put(eureka_url + '/apps/' + app_name + '/' + hostname,
                            headers={'Content-Type': 'application/json'})
    if _response_status_not_ok(response):
        framepy.log.error('[Eureka] Sending heartbeat to cluster failed! Status code {0}'.format(response.status_code))


def _response_status_not_ok(response):
    return response.status_code < 200 or response.status_code >= 300


def _register_heartbeat_service(remote_config_url, app_name, public_hostname):
    def heartbeat_sending_thread():
        while True:
            time.sleep(10)
            _send_heartbeat(remote_config_url, app_name, public_hostname)

    thread = threading.Thread(target=heartbeat_sending_thread)
    thread.daemon = True
    thread.start()


def _get_session_from_cache():
    return _thread_level_cache.fetch_from_cache_or_create_new(SESSION_FIELD, lambda: requests.session())



import inspect


class MethodInspector(object):
    def __init__(self, method):
        self._args = inspect.signature(method).parameters
        self._special_attribute_names = []

        if hasattr(method, '_payload_argument_name'):
            self._special_attribute_names.append(method._payload_argument_name)

    def contains_args(self):
        return len(self._args) > 0

    def check_if_arguments_match_target_method(self, args, kwargs):
        if not self._contains_varargs() and (len(args) + len(kwargs) != len(self._arguments_without_special_ones())):
            raise ValueError('Provided arguments does not match target endpoint')

    def _arguments_without_special_ones(self):
        return list(filter(lambda arg: arg not in ['self'] + self._special_attribute_names, self._args.keys()))

    def _contains_varargs(self):
        for arg_name in self._args:
            if self._args[arg_name].kind == inspect.Parameter.VAR_KEYWORD or \
                            self._args[arg_name].kind == inspect.Parameter.VAR_POSITIONAL:
                return True
        return False



import os
import threading

PID_FIELD_NAME = 'pid'

cache = threading.local()


def fetch_from_cache_or_create_new(field_name, field_creation_procedure):
    if not hasattr(cache, PID_FIELD_NAME):
        setattr(cache, PID_FIELD_NAME, os.getpid())
    if not hasattr(cache, field_name) or getattr(cache, PID_FIELD_NAME) != os.getpid():
        entity = field_creation_procedure()
        setattr(cache, field_name, entity)
        return entity
    else:
        return getattr(cache, field_name)



import random
import requests
import framepy

HTTP_SERVER_ERRORS_BORDER = 500
PROTOCOL = 'http://'


def _perform_operation(context_path, operation, hosts_list, fallback, **kwargs):
    hosts = hosts_list[:]

    if not context_path.startswith('/'):
        context_path = '/' + context_path

    while hosts:
        host = random.choice(hosts)
        try:
            address = PROTOCOL + host + context_path
            response = operation(address, **kwargs)
            if response.status_code < HTTP_SERVER_ERRORS_BORDER:
                return response.json()
        except requests.exceptions.ConnectionError:
            framepy.log.error('Invoking {0} on node {1} failed!'.format(context_path, host))

        hosts = [h for h in hosts if h != host]

    if fallback is not None:
        return fallback(context_path, **kwargs)


def get(context_path, hosts_list, fallback=None, **kwargs):
    return _perform_operation(context_path, requests.get, hosts_list, fallback, **kwargs)


def post(context_path, hosts_list, fallback=None, **kwargs):
    return _perform_operation(context_path, requests.post, hosts_list, fallback, **kwargs)


def put(context_path, hosts_list, fallback=None, **kwargs):
    return _perform_operation(context_path, requests.put, hosts_list, fallback, **kwargs)


def delete(context_path, hosts_list, fallback=None, **kwargs):
    return _perform_operation(context_path, requests.delete, hosts_list, fallback, **kwargs)



import requests
import cherrypy
import configparser
from itertools import chain
from framepy import _utils


def create_configuration(file):
    return _update_with_remote_configuration(_load_properties(file))


def _load_properties(file):
    parser = configparser.RawConfigParser()
    try:
        with open(file, 'r') as f:
            parser.read_file(f)
    except IOError:
        cherrypy.log.error('Cannot open properties file {0}'.format(file))
        raise IOError('Cannot open properties file {0}'.format(file))
    all_properties = [parser.items(section) for section in parser.sections()]
    return {key: value for (key, value) in list(chain(*all_properties))}


def _update_with_remote_configuration(properties):
    remote_config_url = properties.get('remote_config_url')
    app_name = properties.get('app_name')

    if not remote_config_url:
        cherrypy.log.error('Remote config URL not present. Skipping.')
        return properties
    if app_name is None or not app_name:
        cherrypy.log.error('Remote config URL is present but application name was not specified!')
        return properties

    remote_config_url = _utils.normalize_url(remote_config_url)

    remote_properties = {}
    try:
        config_server_response = requests.get(remote_config_url + app_name + '/default')
        if config_server_response.status_code != 200:
            cherrypy.log.error('Error loading remote properties! Status code ' + str(config_server_response.status_code))
            return properties

        sources = [source['source'] for source in config_server_response.json()['propertySources']]
        for source in sources:
            remote_properties.update(source)
    except requests.exceptions.ConnectionError:
        cherrypy.log.error('Error while connecting to remote config server!')
        return properties

    new_properties = properties.copy()
    new_properties.update(remote_properties)
    return new_properties



from . import core

DEFAULT_URL = 'http://localhost'


def normalize_url(url):
    """
    :type url: str
    :rtype: str
    """
    if not url:
        return DEFAULT_URL
    if not url.endswith('/'):
        url += '/'
    return url


def resolve_property_or_report_error(properties, key, log_message, default_value=None):
    """
    :type properties: dict[string, string]
    :type key: str
    :type log_message: str
    :type default_value: *
    """
    value = properties.get(key, None)
    if not value:
        core.log.error(log_message)
        value = default_value
    return value



from framepy import core

annotated_beans = {}


def bean(key):
    def wrapped(potential_bean_class):
        annotated_beans[key] = potential_bean_class
        return potential_bean_class
    return wrapped


class BeansInitializer(object):
    def __init__(self):
        self.initial_mappings = []
        for key, bean in annotated_beans.items():
            self.initial_mappings.append(core.Mapping(bean(), key))
        self.all_beans = {bean.path: bean.bean for bean in self.initial_mappings}

    def initialize_all(self, context):
        for key, bean in self.all_beans.items():
            self.initialize_bean(key, bean, context)

    def initialize_bean(self, target_bean_name, target_bean, context):
        for key, bean in self.all_beans.items():
            if bean != target_bean:
                property = 'set_' + key
                if hasattr(target_bean, property):
                    getattr(target_bean, property)(bean)

        try:
            target_bean.initialize(context)
        except Exception as e:
            raise BeanInitializationException("Cannot initialize bean '{0}'".format(target_bean_name), e)


class BeanInitializationException(Exception):
    pass



import cherrypy
import logging
from cherrypy._cplogging import LogManager

ERROR_LOG_FORMATTER = logging.Formatter('[%(asctime)s] %(levelname)s %(filename)s:%(funcName)s: %(message)s')
ACCESS_LOG_FORMATTER = '{h} {l} {u} "{r}" {s} {b} "{f}" "{a}"'


def setup_logging(logger):
    _hack_cherrypy_logs()

    application_log_handler = _setup_application_log()
    access_log_handler = _setup_access_log_handler()

    logger.addHandler(application_log_handler)

    cherrypy.log.error_log.handlers = []
    cherrypy.log.error_log.addHandler(application_log_handler)
    cherrypy.log.access_log.handlers = []
    cherrypy.log.access_log.addHandler(access_log_handler)


def _hack_cherrypy_logs():
    LogManager.time = lambda *_: ''
    LogManager.access_log_format = ACCESS_LOG_FORMATTER


def _setup_access_log_handler():
    access_log_handler = logging.StreamHandler()
    access_log_handler.setLevel(logging.INFO)
    access_log_handler.setFormatter(ERROR_LOG_FORMATTER)
    return access_log_handler


def _setup_application_log():
    application_log_handler = logging.StreamHandler()
    application_log_handler.setLevel(logging.INFO)
    application_log_handler.setFormatter(ERROR_LOG_FORMATTER)
    return application_log_handler



import collections
from framepy import logs
from framepy import beans
from framepy import web
from framepy import _configuration
import logging
import cherrypy
import pkgutil

DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 8000
Mapping = collections.namedtuple('Mapping', ['bean', 'path'])

log = logging.getLogger('framepy_logger')


class Context(object):
    def __init__(self, entries):
        self.__dict__.update(entries)


class BaseBean(object):
    def __init__(self):
        self.context = None

    def initialize(self, context):
        self.context = context


def _setup_server_config(load_properties):
    cherrypy.config.update(
        {
            'server.socket_port': load_properties.get('server_port', DEFAULT_PORT),
            'server.socket_host': load_properties.get('server_host', DEFAULT_HOST),
            'tools.encode.on': True,
            'tools.encode.encoding': 'utf-8'
        }
    )


def _setup_modules(loaded_properties, modules, kwargs):
    beans = {}
    for module in modules:
        module.before_setup(loaded_properties, kwargs, beans)

    return Context(beans)


def _after_setup(context, modules, kwargs, properties, beans_initializer):
    beans_initializer.initialize_all(context)
    for module in modules:
        module.after_setup(properties, kwargs, context, beans_initializer)


def _create_context(beans_initializer, kwargs, loaded_properties, modules):
    context = _setup_modules(loaded_properties, modules, kwargs)
    _after_setup(context, modules, kwargs, loaded_properties, beans_initializer)


def _finish_system_initialization(loaded_properties):
    _setup_server_config(loaded_properties)
    logs.setup_logging(log)


def scan_packages(packages_filter=lambda _: True):
    for modname in (modname for importer, modname, ispkg in pkgutil.walk_packages('.')
                    if '.' in modname and packages_filter(modname)):
        __import__(modname)


def init_context(properties_file,
                 modules=(),
                 **kwargs):
    beans_initializer = beans.BeansInitializer()
    modules = (web.Module(),) + modules

    properties = _configuration.create_configuration(properties_file)

    _finish_system_initialization(properties)
    _create_context(beans_initializer, kwargs, properties, modules)

    return cherrypy.tree


def start_standalone_application():
    cherrypy.engine.start()
    cherrypy.engine.block()



from . import modules


class Module(modules.Module):
    def before_setup(self, properties, arguments, beans):
        aws_access_key = properties['aws_access_key']
        aws_access_secret_key = properties['aws_access_secret_key']
        aws_region = properties['aws_region']
        endpoint_url = self._filter_missing_endpoint(properties.get('aws_endpoint_url', ''))

        beans['aws_credentials'] = {
            'aws_access_key_id': aws_access_key,
            'aws_secret_access_key': aws_access_secret_key,
            'region_name': aws_region,
            'endpoint_url': endpoint_url
        }

    def after_setup(self, properties, arguments, context, beans_initializer):
        pass

    @staticmethod
    def _filter_missing_endpoint(endpoint_url):
        return endpoint_url if len(endpoint_url) > 0 else None


def get_credentials(context):
    return context.aws_credentials
